#!/bin/bash

# --------------------------------
export cGitProjVersion cGetOrigin cGetTopDir cGetCurBranch cRemoteProjFile cPID

export PWD gErr gSavedOptE gGetConfigOrigin gGetConfigValue gResponse

export gpAuto gpAutoMove gpBin gpCheckFileNames gpCheckForBigFiles \
    gpCheckInRaw gpCheckWhitespace gpCmdName gpDebug gpDoc gpFacility \
    gpGitFlow gpHookVerbose gpLocalRawDir gpLocalStatus gpLocalTopDir \
    gpMaxLoop gpMaxSize gpMountDir gpPath gpPreCommitEnabled gpProjName \
    gpProjStatus gpRemoteStatus gpSysLog gpVer gpVerbose gpYesNo

# --------------------------------
fComGitProjInternalDoc()
{
    return
    # This should be the first "internal-pod"
    cat <<\EOF >/dev/null
=internal-pod

=internal-for text ========================================

=internal-for html <hr/>

=internal-head1 gitproj-com.inc Internal Documentation

=internal-for html <hr/>

=internal-head2 Common Script Functions

=internal-cut

EOF
} # fGitProjComInternalDoc

# ========================================
# Common Functions

# --------------------
fComIntroText()
{
    local tSrc=${BASH_SOURCE##*/}

    cat <<EOF | fComFmt

======================================================================

You now have a local git repository project (${gpProjName}) setup at
${gpLocalTopDir} with support for managing large binary files (larger
than $gpMaxSize).

~/.gitconfig the '[gitproj "config"]' and '[gitproj "hooks"]'
sections, which will defined the initial gitproj variables for new
projects.  The global variables can be overridden by a project's local
config definitions (in .git/config).

EOF
    return 0
} # fComIntroText

# --------------------
fComConfigCopy()
{
    local pForce=0
    local pSource=""
    local pDest=""
    local pIncPat=""
    local pExcPat="@@"
    local tSrc=${BASH_SOURCE##*/}
    local tVerbose=""
    declare -l tVar
    local tValue
    local tFound
    local OPTIND=1
    local OPTARG
    local tArg
    local tDestDir

    if [ $gpVerbose -ge 2 ]; then
        tVerbose='-v'
    fi

    # Get args and validate
    while getopts :fs:d:i:e: tArg; do
        # shellcheck disable=SC2213,SC2220
        case $tArg in
            f) pForce=1 ;;
            s) pSource="${OPTARG}" ;;
            d) pDest="${OPTARG}" ;;
            i) pIncPat="${OPTARG}" ;;
            e) pExcPat="${OPTARG}" ;;
            :)
                fError -n -i -m "fComConfigCopy: Value required for option: $OPTARG" -l $tSrc:$LINENO
                return 1
                ;;
            \?)
                fError -n -i -m "ComConfigCopy: Unknown option: $OPTARG" -l $tSrc:$LINENO
                return 1
                ;;
        esac
    done

    if [[ -z "$pSource" || ! -r $pSource ]]; then
        fError -n -i -m "fComConfigCopy: -s $pSource is missing or not readable" -l $tSrc:$LINENO
        return 1
    fi
    if [ -z "$pDest" ]; then
        fError -n -i -m "fComConfigCopy: -d $pDest is missing or not writable" -l $tSrc:$LINENO
        return 1
    fi

    tDestDir=${pDest%/*}
    if [ "$tDestDir" != "$pDest" ]; then
        if [ ! -d $tDestDir ]; then
            fError -n -i -m "$tDestDir directory does not exist" -l $tSrc:$LINENO
            return 1
        fi
    fi

    touch $pDest
    if [ ! -w $pDest ]; then
        fError -n -i -m "-d $pDest file is not writable" -l $tSrc:$LINENO
        return 1
    fi
    if ! cp $tVerbose --backup=t $pDest $pDest.bak; then
        fError -n -i -m "Problem backing up -d $pDest file" -l $tSrc:$LINENO
        return 1
    fi

    git config -f $pSource --list | \
      grep -E "$pIncPat" | grep -Ev "$pExcPat" | \
      while read tSrcConfig; do
        tVar=${tSrcConfig%=*}
        tValue="${tSrcConfig#*=}"
        if [ $pForce -ne 0 ]; then
            git config -f $pDest $tVar "$tValue"
            if [ $? -ne 0 ]; then
                fError -i -n "git config error" -l $tSrc:$LINENO
                fLog -f -p warning -m "git config -f $pDest $tVar $tValue" -l $tSrc:$LINENO
            fi
            continue
        fi
        if git config -f $pDest --get $tVar >/dev/null 2>&1; then
            continue
        fi
        git config -f $pDest $tVar "$tValue"
        if [ $? -ne 0 ]; then
            fError -i -n "git config error" -l $tSrc:$LINENO
        fi
    done
    if diff -q $pDest $pDest.bak >/dev/null 2>&1; then
        rm -f $pDest.bak
    fi

    return 0
    # ----------------------
    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComConfigCopy

=internal-head4 Synopsis

    fComConfigCopy [-f] -s pSource -d pDest [-i pInclPat] [-e pExclPat]

=internal-head4 Description

Copy values from git config file pSource to pDest. If -f (force), the pSource
values will override the pDest values, otherwise the pSource value will
only be copied to pDest if it does not exist there.

The -i pIncPat is a "grep" pattern that only selects the variable
names that match the pattern. For example:
S<-i 'gitproj\.config\.|gitflow\.'>

The -e pExcPat is a "grep" pattern that removes variable names that
match the pattern. For example: S<-e 'gitproj.config.remote-raw-origin'>
pExcPat is applied after pIncPat.

pSource must exist and be readable. If pDest exists, must be writable.
If pDest does not exist then it will be created with the specified
include and exclude patterns.

A backup copy of pDest is made with this command:
S<cp -backup=t $pDest $pDest.bak>

=internal-cut
EOF
} # fComConfigCopy

# --------------------
fComConfigUpdateLocal()
{
    local tSrc=${BASH_SOURCE##*/}
    local tTopDir
    local tExclude="local-status|local-host|remote-status|remote-raw-origin"

    if ! fComMustBeInProjRepo $PWD >/dev/null 2>&1; then
        fError -i -n -m "fComConfigUpdateLocal: Missing: $tTopDir/.gitproj" -l $tSrc:$LINENO
        return 1
    fi

    tTopDir=$($cGetTopDir)
    if [ -f $tTopDir/.gitproj ] && [ $tTopDir/.git/config -ot $tTopDir/.gitproj ]; then
        return 0
    fi

    fComConfigCopy -f -s $tTopDir/.git/config -d $tTopDir/.gitproj \
        -i 'gitproj\.' -e "$tExclude"

    # Force TBD values?
    ## fComConfigCopy -f -s $gpDoc/config/gitconfig -d -d $tTopDir/.gitproj \
    ##    -i $tExclude

    git add $tTopDir/.gitproj
    git commit -m "Updated .gitproj" $tTopDir/.gitproj
    return 0

    # or just be noisy?
    # if ! git ls-files --error-unmatch $tTopDir/.gitproj >/dev/null 2>&1; then
    #     fError -n -m "You must version $tTopDir/.gitproj" -l $tSrc:$LINENO
    # fi
} # fComConfigUpdateLocal

# --------------------
fComFirstTimeSet()
{
    local tSrc=${BASH_SOURCE##*/}
    local tInclude
    local tUserName
    local tUserEmail
    local tVerbose

    if [ $gpVerbose -ge 2 ]; then
        tVerbose='-v'
    fi

    if [ ! -f $HOME/.gitconfig ]; then
        fLog -p info -m "Create: $HOME/.gitconfig" -l $tSrc:$LINENO
        cp $tVerbose --backup=t $gpDoc/config/gitconfig $HOME/.gitconfig
        return 0
    fi

    tUserName=$(fComGetConfig -g -k "user.name")
    tUserEmail=$(fComGetConfig -g -k "user.email")
    if [ -z "$tUserName" ] || [ -z "$tUserEmail" ]; then
        fError -m "user.name and user.email must be in your ~/.gitconfig file. See $gpDoc/config/gitconfig for some other useful settings." -l $tSrc:$LINENO
    fi

    fComConfigCopy -s $gpDoc/config/gitconfig -d $HOME/.gitconfig
    return 0
} # fComFirstTimeSet

# -------------------
fComSetGlobals()
{
    local tSrc=${BASH_SOURCE##*/}

    # -------------------
    # Set current directory location in PWD, because with cron jobs PWD
    # is not set.
    if [ -z "$PWD" ]; then
        PWD=$(pwd)
    fi

    # -------------------
    # Define the location of the com script if not already set
    if [ -z "$gpBin" ]; then
        gpBin=${0%/*}
        if [ "$gpBin" = "." ]; then
            gpBin=$PWD
        fi
        cd $gpBin >/dev/null 2>&1
        gpBin=$PWD
        cd - >/dev/null 2>&1
    fi

    # -------------------
    # Constants

    gpDoc=${gpDoc:-/usr/share/doc/git-proj}
    if [ ! -d $gpDoc ]; then
        gpDoc=$gpBin/../doc
    fi
    if [ ! -d $gpDoc ]; then
        fError -m "Error: Cannot find gitproj doc directory: $gpDoc" -l $tSrc:$LINENO
    fi
    cd $gpDoc >/dev/null 2>&1
    gpDoc=$PWD
    cd - >/dev/null 2>&1

    cGitProjVersion=$(cat $gpDoc/VERSION)

    # Include in ~/.gitconfig (Use -g to set config vars to this)
    # Note: ~/.gitconfig does not work all the time, so use:
    # $HOME/.gitconfig or use --global

    # Include in PROJ/.gitproj (Use -L to set config vars to this)
    # Include in PROJ/.git/config (Use -l to set config vars to this)

    cGetOrigin='git config --get remote.origin.url'
    # Use: tOrigin=$($cGetOrigin)
    #      if $cGetOrigin >/dev/null 2>&1; then echo "In a git dir"; fi

    cGetTopDir='git rev-parse --show-toplevel'
    # Use: tTop=$($cGetTopDir); cd $tTop

    cGetCurBranch='git rev-parse --abbrev-ref HEAD'
    # Use: tBranch=$($cGetCurBranch)

    cRemoteProjFile=.remote.proj
    # This file will contain cGetOrigin value. Put in remote raw dir

    gpMaxLoop=${gpMaxLoop:-20}
    # Used internally to limit number of invalid responses to prompts.
    # For testing this is usually reduced.

    # -------------------
    # Setup log variables
    # -x
    gpDebug=${gpDebug:-0}

    # -v
    gpVerbose=${gpVerbose:-2}

    # -l
    gpSysLog=${gpSysLog:-false}
    gpFacility=${gpFacility:-user}

    gErr=0

    return 0

    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComSetGlobals

Set initial values for all of the globals use by this script. The ones
that begin with "gp" can usually be overridden by setting them before
the script is run.

=internal-cut

EOF
} # fComSetGlobals

# --------------------
fComPreProjSetGlobals()
{
    local tSrc=${BASH_SOURCE##*/}
    local tTrim

    fComFirstTimeSet

    gpVer=$cGitProjVersion

    gpSysLog=$(fComGetConfig -k "gitproj.config.syslog" -d "false")
    gpFacility=$(fComGetConfig -k "gitproj.config.facility" -d user)

    gpVerbose=$(fComGetConfig -k "gitproj.config.verbose" -i -d 2)

    gpAuto=0
    gpLocalTopDir=$PWD
    gpProjName=$(fComGetConfig -k "gitproj.config.name" -d "TBD")
    if [ "$gpProjName" = "TBD" ]; then
        gpProjName="${gpLocalTopDir##*/}"
    fi

    gpLocalRawDir=$gpLocalTopDir/raw
    gpRemoteMinSpace=$(fComGetConfig -i -k "gitproj.config.remote-min-space" -d "2147483648") #2g

    gpAutoMove="true"
    gpGitFlow="false"
    gpGitFlowPkg=$(fComGetConfig -k "gitproj.config.git-flow-pkg" -d "git-flow")

    gpProjStatus=$(fComGetConfig -k "gitproj.config.proj-status" -d "installed")
    gpLocalStatus=$(fComGetConfig -k "gitproj.config.local-status" -d "not-defined")
    gpRemoteStatus=$(fComGetConfig -k "gitproj.config.remote-status" -d "not-defined")

    gpMountDir=TBD
    gpRemoteRawOrigin=TBD

    # Hook section
    gpHookVerbose=$(fComGetConfig -k "gitproj.hook.verbose" -d true)
    gpPreCommitEnabled=$(fComGetConfig -k "gitproj.hook.pre-commit-enabled" -d true)
    gpCheckFileNames=$(fComGetConfig -k "gitproj.hook.check-file-names" -d true)
    gpCheckInRaw=$(fComGetConfig -k "gitproj.hook.check-in-raw" -d true)
    gpCheckWhitespace=$(fComGetConfig -k "gitproj.hook.check-whitespace" -d true)
    gpCheckForBigFiles=$(fComGetConfig -k "gitproj.hook.check-for-big-files" -d true)
    gpMaxSize=$(fComGetConfig -k "gitproj.hook.binary-file-size" -d "10k")

    # These will be redefined later
    #    gpLocalTopDir
    #    gpProjName, gitproj.config.name
    #    gpLocalRawDir=$gpLocalTopDir/raw
    #    gpGitFlow="false"
    #    gpLocalStatus, gitproj.config.local-status
    #    gpRemoteStatus, gitproj.config.remote-status
    #    gpCheckFileNames=true
    #    gpCheckForBigFiles=true
    #    gpPreCommitEnabled=true
    #    gpMountDir=TBD
    #    gpRemoteRawOrigin=TBD

    return 0
} # fComPreProjSetGlobals

# --------------------------------
fComCheckDeps()
{
    local pRequired="$1"
    local pOptional="$2"
    local tSrc=${BASH_SOURCE##*/}
    local tProg
    local tErr=0
    local tComReq='logger pod2text pod2usage rsync'
    local tComOpt='pod2html pod2man pod2markdown tidy shunit2.1 gawk sed tr fmt'

    for tProg in $pOptional $tComOpt; do
        if ! command -v $tProg >/dev/null 2>&1; then
            echo "Optional: Missing $tProg"
            tErr=1
        fi
    done

    for tProg in $pRequired $tComReq; do
        if ! command -v $tProg >/dev/null 2>&1; then
            echo "Required: Missing $tProg"
            tErr=2
        fi
    done

    if [ $tErr -eq 2 ]; then
        fError -m "Missing one or more required programs." -l $tSrc:$LINENO
    fi
    if [ $tErr -eq 1 ]; then
        fLog -p warning -m "Missing some some optional programs." -l $LINENO
    fi
    return 0

    cat <<EOF >/dev/null
=internal-pod

=internal-for text ========================================

=internal-for html <hr/>

=internal-head3 fComCheckDeps "pRequired List" "pOptional List"

Check for required and optional programs or scripts used by this script.
If any required programs are missing, exit the script.

=internal-cut
EOF
} # fComCheckDeps

# --------------------------------
fComInternalDoc()
{
    local tSrc=${BASH_SOURCE##*/}

    awk '
        /^=internal-pod/,/^=internal-cut/ {
            sub(/^=internal-/,"=");
            print $0;
        }
    '
    return 0

    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComInternalDoc

This function collects all of the "internal-pod" documentation from
stdin and it outputs to stdout.

=internal-cut
EOF
} # fComInternalDoc

# --------------------------------
fComUsage()
{
    local pFileList=""
    local pStyle="long"
    local pInt=0
    local pTitle="$gpCmdName Usage"
    local tSrc=${BASH_SOURCE##*/}
    local tProg=""
    local tTidy='tidy -q -i -w 78 -asxhtml --break-before-br yes --indent-attributes yes --indent-spaces 2 --tidy-mark no --vertical-space no'
    local tHtmlOpt='--cachedir=/tmp --index'
    local OPTIND=1
    local OPTARG
    local tArg

    while getopts :f:s:t:i tArg; do
        case $tArg in
            f) pFileList="$pFileList $OPTARG" ;;
            s) pStyle="$OPTARG" ;;
            t) pTitle="$OPTARG" ;;
            i) pInt=1 ;;
            :) fError -i -m "fComUsage: Value required for option: -$OPTARG" -l $tSrc:$LINENO ;;
            \?) fError -i "fComUsage: Unknown option: $OPTARG" -l $tSrc:$LINENO ;;
        esac
    done
    shift $((OPTIND - 1))
    if [ $# -ne 0 ]; then
        fError -i -m "fComUsage: Unknown option: $*" -l $tSrc:$LINENO
    fi

    case $pStyle in
        short | usage)
            tProg=pod2usage
            if [ $pInt -ne 0 ]; then
                # This fails, don't know why
                # tProg='pod2text | head -n 24'
                tProg=pod2text
            fi
            ;;
        man)
            tProg=pod2man
            if [ $pInt -ne 0 ]; then
                tProg=pod2text
            fi
            ;;
        long | text)
            tProg=pod2text
            ;;
        html)
            tProg=pod2html
            ;;
        md)
            tProg=pod2markdown
            ;;
        *)
            tProg=pod2text
            fLog -p warning -m "Unknown style: '$pStyle' so using text" -l $tSrc:$LINENO
            ;;
    esac

    # Validate
    if ! command -v $tProg >/dev/null 2>&1; then
        fLog -p warning -m "Could not find $tProg so using text" -l $tSrc:$LINENO
        tProg=pod2text
    fi

    # shellcheck disable=SC2116
    pFileList="$(echo $pFileList)"
    if [ -z "$pFileList" ]; then
        fError -i -m "fComUsage missing pFileList" -l $tSrc:$LINENO
    fi
    for tFile in $pFileList; do
        if [ ! -r $tFile ]; then
            fError -i -m "fComUsage cannot read: $tFile" -l $tSrc:$LINENO
        fi
    done

    if [ $pInt -eq 0 ]; then
        if [ "$pStyle" = "html" ]; then
            # shellcheck disable=SC2002
            cat $pFileList |
                $tProg --title="$pTitle" $tHtmlOpt |
                $tTidy 2>/dev/null | more
        else
            if [ "$pStyle" = "short" ]; then
                # shellcheck disable=SC2002
                cat $pFileList | pod2text | head -n 10
            else
                # shellcheck disable=SC2002
                cat $pFileList | $tProg | more
            fi
        fi
    else
        if [ "$pStyle" = "html" ]; then
            # shellcheck disable=SC2002
            cat $pFileList | fComInternalDoc |
                $tProg --title="$pTitle" $tHtmlOpt |
                $tTidy 2>/dev/null | more
        else
            # shellcheck disable=SC2002
            cat $pFileList | fComInternalDoc | $tProg | more
        fi
    fi

    gErr=1
    exit 1

    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComUsage -f pFileList -s pStyle [-t pTitle] [-i] [-a]

=internal-over 4

=internal-item B<-f pFileList> - list of file names

=internal-item B<-s pStyle> - output style

 short|usage - usage only (does not work with -i or -a)
 man         - all, man format (does not work with -i or -a)
 long|text   - all text format
 html        - all, html format (see -t)
 md          - all, markdown format

=internal-item B<-t> - title for HTML style

=internal-item B<-i> - internal doc only (see fComInternalDoc)

=internal-back

=internal-cut
EOF
} # fComUsage

# ------------------
fComStackTrace()
{
    local i
    local tSrc
    local tLine
    local tName
    local tOutput

    # Credit: https://gist.github.com/ahendrix/7030300
    if [ ${#FUNCNAME[@]} -le 2 ]; then
        return 0
    fi

    if [ "${gpSysLog:-false}" = "true" ]; then
        tOutput="logger -s -t ${gpCmdName:-fComStackTrace} -p ${gpFacility:-user}.debug"
    else
        tOutput="cat"
    fi

    tSrc=${BASH_SOURCE[0]}
    tSrc=${tSrc##*/}
    tLine=${BASH_LINENO[0]}
    tName="${FUNCNAME[1]}()"
    echo "Stack trace at: ${tSrc}:${tLine} ${tName}" | $tOutput
    for ((i = 1; i < ${#FUNCNAME[@]} - 1; i++)); do
        tSrc=${BASH_SOURCE[$i + 1]}
        tSrc=${tSrc##*/}
        tLine=${BASH_LINENO[$i]}
        tName="${FUNCNAME[$i]}()"
        echo " $i: ${tSrc}:${tLine} ${tName}" | $tOutput
    done

    return 0
} # fComStackTrace

# --------------------
fComSelect()
{
    PS3="$1"
    local pList="$2"
    local pHelp="$3"
    local tSrc=${BASH_SOURCE##*/}
    local tResponse
    local tStatus=1
    local tMaxLoop=30
    export gResponse

    # Loop until a valid selection, or QUIT or OTHER
    select tResponse in QUIT HELP $pList; do
        let --tMaxLoop
        if [ $tMaxLoop -le 0 ]; then
            fLog -f -p warning -m "Too many responses." -l $tSrc:$LINENO
            tResponse="QUIT"
            break
        fi
        echo
        case $tResponse in
            QUIT)
                fLog -f -p warning -m "Quitting." -l $tSrc:$LINENO
                break
                ;;
            HELP)
                echo 1>&2
                echo "$pHelp" 1>&2
                echo 1>&2
                continue
                ;;
        esac
        if [ -n "$tResponse" ]; then
            tStatus=0
            break
        fi
        if [ -z "$tResponse" ]; then
            fLog -f -p warning -m "Invalid selection" -l $tSrc:$LINENO
            continue
        fi
    done

    PS3=""
    gResponse=$tResponse
    return $tStatus
} # fComSelect

# ------------------
fComYesNo()
{
    local pPrompt="$1"
    local tSrc=${BASH_SOURCE##*/}
    declare -l tResult

    if [ ${gpAuto:-0} -ne 0 ]; then
        if [ "$gpYesNo" = "y" ]; then
            gResponse="y"
            return 0
        fi
        if [ "$gpYesNo" = "n" ]; then
            gResponse="n"
            return 1
        fi
        gResponse="n"
        return 1
    fi
    echo -n "$pPrompt [y/n]? "
    read -r tResult
    if [ "${tResult::1}" = "y" ]; then
        gResponse="y"
        return 0
    fi
    gResponse="n"
    return 1

    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComYesNo pPrompt

If gpYesNo = "y", set gResponse to "y", and return 0.

If gpYesNo = "n", set gResponse to "n", and return 1.

If gpAuto is 0, output the pPrompt, then "read" the response.

If the response begins with a "y", then that is the same as gpYesNo = "y"

If the response is not a "y", then that is the same as gpYesNo = "n"

=internal-cut
EOF
} # fComYesNo

# ------------------
fComMenu()
{
    local pTitle=$1
    local pMenu=$2
    local pActn=$3
    local tSrc=${BASH_SOURCE##*/}
    declare -g gMenuTitle
    declare -a tMenu
    local tMaxLoop=$gpMaxLoop
    local tActn
    local tSize
    local tItem
    local tPrompt
    local tSelect
    local tCmd
    local tFun
    local tIFS

    tSize=$(eval echo "\${#${pMenu}[@]}")
    let --tSize
    tPrompt="$(eval echo $(echo \${${pMenu}[0]}))"
    while true; do
        let --tMaxLoop
        if [ $tMaxLoop -le 0 ]; then
            fLog -f -p warning -m "Too many responses." -l $tSrc:$LINENO
            break
        fi
        gMenuTitle=$pTitle
        tSelected=$(echo $pTitle | sed 's/{[^}]*} //g')
        echo
        echo "Current selections:"
        tIFS="$IFS"
        IFS=';'
        printf '\t%s\n' $tSelected
        IFS="$tIFS"
        for ((tItem = 1 ; tItem <= tSize ; tItem++)); do
            tMenu[$tItem]="$(eval echo $(echo \${${pMenu}[$tItem]}))"
            echo -e "  $tItem)\n      ${tMenu[$tItem]}" | \
                sed 's/{[^}]*} //g' | fmt -t
        done
        echo -n "$tPrompt "
        read tSelect
        if [ $gpDebug -ne 0 ]; then
            echo "$tSelect"
        fi
        if [ -z "$tSelect" ]; then
            continue
        fi
        echo "$tSelect" | grep -q '[0-9]'
        if [ $? -ne 0 ]; then
            continue
        fi
        if [[ $tSelect -lt 1 || $tSelect -gt $tSize ]]; then
            continue
        fi
        tActn="$(eval echo $(echo \${${pActn}[$tSelect]}))"
        if [ -z "$tActn" ]; then
            fError -i -n -m "Problem with action" -l $tSrc:$LINENO
            continue
        fi
        tCmd=${tActn%% *}
        tFun=${tActn#* }
        case $tCmd in
            help)
                fComMenuHelp $pActn
                tMaxLoop=$gpMaxLoop
                ;;
            func)
                eval $tFun
                tMaxLoop=$gpMaxLoop
                ;;
            menu)
                fComMenu "${pTitle};${tMenu[$tSelect]}" ${tFun}
                tMaxLoop=$gpMaxLoop
                ;;
            back)
                return 0
                ;;
            *)
                echo
                fError -i -n -m "Problem with action" -l $tSrc:$LINENO
                ;;
        esac
        continue
    done
    return 1
    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComMenu pTitle pMenu pActn

pTitle - string

    If the Actn is a "menu", the selected menu's text will be append
    to the current pTitle, separated with ';'. And gMenuTitle will be
    set to that combined title.

    When the pTitle is displayed, the ';' is replaced with a newline.
    Also all text that matches '{.*} ' will be removed. This is done
    so that the text between '{}' can be used internally, to parse
    what was selected.

    Examples:

    pTitle="First title"
    gMenu1[0]="Select the text style:"
    gMenu1[1]="{code} Select text for Code1"
    gActn1[1]="menu gMenu2 gActn2"
    gMenu1[2]="{code2} Select alternate text"
    gActn1[2]="menu gMenu2 gActn2"
    gMenu1[3]="Quit"
    gActn1[3]="func exit 1"
    This will be displayed as:

    Current selections:
            First title
    1) Select text for Code1
    2) Select alternate text
    3) Quit
    Select the text style:

    User selects menu item 1, and the next menu will be
    pTitle="First title;code1: Select text for Code1"
    gMenu2[0]="Select the type:"
    gMenu2[1]="{file} Change file"
    gActn2[1]="menu gYesNoMenu gYesNoActn
    gMenu2[2]="{section} Change section"
    gActn2[2]="menu gYesNoMenu gYesNoActn
    gMenu2[3]="Back"
    gActn2[3]="back"
    This will be desplayed as:

    Current selections:
            First title
            Select text for Code1
    1) Change file
    2) Change section
    3) Back
    Select the type:

    User selects menu item 2, and the next menu will be
    pTitle="First title;code1: Select text for Code1;section: Change section"
    gYesNoMenu[0]="Are the above selections OK?"
    gYesNoMenu[1]="Yes, continue"
    gYesNoActn[1]="func fProcess \$gMenuTitle"
    gYesNoMenu[2]="No, exit"
    gYesNoActn[2]="func exit 1"
    gYesNoMenu[3]="Back"
    gYesNoActn[3]="back"

    This will be desplayed as:

    Current selections:
            First title
            Select text for Code1
            Change section
    1) Yes, continue
    2) No, exit
    3) Back
    Are the above selections OK?

    If the user selects 2, then then fProcess will be called:

    fProcess First title;{code1} Select text for Code1;{section} Change section

pMenu - the name of an array of menu items

    pMenu[0] - prompt for the menu
    pMenu[1..Size] - text is piped through "fmt -t" (after prepending "N) ")

    Also text that matches '{.*}' will be removed.

    Example:
    pMenu[1]="code1: Select text for Code1"

pActn - the name of an array of actions for a menu item with the same index.

    pActn[0] - help text for the menu
    pActn[1..Size]  - first word is a command, followed by options

pActn Commands

    help - send the pActn array name to fComMenuHelp, it will output
        pActn[0].

    func pFuncName [pArgs...]- call pFuncName passing pArgs. The
        \$gMenuTitle can be passed to the function, so that the list of
        what was selected can be parsed. Look for ';text: ' for the
        selection codes.

    menu pMenu pActn - call fComMenu passing the specified array names
        The pTitle will be the text from the calling menu item.

    back - return to the calling menu (or the first calling program).

Examples

    See test/manual-test-menu.sh
    See testComMenu() in test/test-com.sh
    See fConfigMenu() in git-core/gitproj-config.inc

=internal-cut
EOF
} # fComMenu

# ------------------
fComMenuHelp()
{
    local pHelp=$1
    local tSrc=${BASH_SOURCE##*/}
    local tHelp

    tHelp="$(eval echo -e "$(echo -e "\${${pHelp}[0]}")")"
    echo
    echo '--------------------'
    echo -e "$tHelp" | fmt -t | grep .
    echo '--------------------'
} # fComMenuHelp

# ------------------
fComFmtLog()
{
    local pLevel=$1
    local pMsg="$2"
    local pLine=$3
    local pErr=$4
    local pForce=$5
    local tSrc=${BASH_SOURCE##*/}
    local tDebugLevel
    local tLevel
    local tLine
    local tErr

    # Use this to turn PID on/off in logs (hard-coded)
    local tPID_Flag=0
    local tPID=""
    local tPID_Opt=""

    # Set any missing globals
    gpSysLog=${gpSysLog:-false}
    gpFacility=${gpFacility:-user}
    gpVerbose=${gpVerbose:-0}
    gpDebug=${gpDebug:-0}

    tLevel=$pLevel
    case $pLevel in
        emerg | emergency)
            pLevel=emerg
            tLevel=$pLevel
            ;;
        alert)
            :
            ;;
        crit | critical)
            pLevel=crit
            tLevel=$pLevel
            ;;
        err | error)
            pLevel=err
            tLevel=$pLevel
            ;;
        warning | warn)
            if [ $pForce -eq 0 ] && [ $gpVerbose -eq 0 ]; then
                return
            fi
            pLevel=warning
            tLevel=$pLevel
            ;;
        notice)
            if [ $pForce -eq 0 ] && [ $gpVerbose -lt 2 ]; then
                return
            fi
            ;;
        info)
            if [ $pForce -eq 0 ] && [ $gpVerbose -lt 3 ]; then
                return
            fi
            ;;
        debug | debug-*)
            tLevel=debug
            if [ $gpDebug -eq 0 ]; then
                return
            fi
            if [ $gpDebug -ne 0 ] && [ "${pLevel%%-*}" != "$pLevel" ]; then
                tDebugLevel=${pLevel##*-}
                if [ $tDebugLevel -gt $gpDebug ]; then
                    return
                fi
            fi
            ;;
        *)
            pMsg="Internal: (Unknown: $pLevel [$tSrc:$LINENO]) $pMsg"
            pLevel=err
            tLevel=$pLevel
            ;;
    esac

    # LineNo format
    tLine=""
    if [ -n "$pLine" ]; then
        tLine="[$pLine]"
    fi

    # Err format
    tErr=""
    if [ -n "$pErr" ]; then
        tErr="($pErr)"
    fi

    if [ $tPID_Flag -eq 1 ]; then
        tPID="[$tPID]"
        tPID_Opt="-i"
    fi

    # Output
    if [ "$gpSysLog" = "false" ]; then
        echo "${gpCmdName}$tPID $pLevel: $pMsg $tLine$tErr" 1>&2
    else
        logger -s $tPID_Opt -t $gpCmdName -p $gpFacility.$tLevel "$pLevel: $pMsg $tLine$tErr"
    fi
    return 0

    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComFmtLog pLevel "pMsg" pLine pErr

This function formats and outputs a consistent log message output.
See: fLog and fError.

=internal-cut
EOF
} # fComFmtLog

# ------------------
fLog()
{
    local pLevel="info"
    local pMsg=""
    local pLine=""
    local pErr=""
    local pInt=0
    local pForce=0
    local tSrc=${BASH_SOURCE##*/}
    local OPTARG=1
    local OPTIND
    local tArg

    while getopts :m:p:l:e:if tArg; do
        case $tArg in
            m) pMsg="${OPTARG}" ;;
            l) pLine="${OPTARG}" ;;
            p) pLevel="${OPTARG}" ;;
            e) pErr="${OPTARG}" ;;
            i) pInt=1 ;;
            f) pForce=1 ;;
            :) echo "Internal: Value required for option: -$OPTARG [$tSrc:$LINENO]" ;;
            \?) echo "Internal: Unknown option: $OPTARG [$tSrc:$LINENO]" ;;
        esac
    done
    shift $((OPTIND - 1))
    if [ $# -ne 0 ]; then
        shift $#
    fi

    if [ $pInt -eq 1 ]; then
        pMsg="Internal: $pMsg"
    fi

    fComFmtLog "$pLevel" "$pMsg" "$pLine" "$pErr" "$pForce"
    return 0

    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fLog -m pMsg [-p pLevel] [-l $LINENO] [-e pErr] [-f]

pLevel - emerg alert crit err warning notice info debug debug-N

 emerg, alert, crit, and err levels will always be output.

 emerg, alert, crit - exit gErr (or 1 if gErr=0, or not -e)
 err - return gErr (or 1 if gErr=0, or not -e)
 all other levels - return 0

 (fError defaults to "crit". -n changes it too "err")
 (-i adds "Internal: to message, an it calls fComStackTrace)

 -q   set gpVerbose=0
 -v   set gpVerbose=2 (default)
 -V N set gpVerbose=N (0..4)
 gitproj.config.verbose=N set gpVerbose=N

 if gpVerbose = 0, don't output warning, notice or info
 if gpVerbose = 1, don't output notice, info
 if gpVerbose = 2, don't output info (default verbose level)
 if gpVerbose >= 3, output all non-debug levels
 if -f, ignore the gpVerbose setting

 -x - set gpDebug=1
 -xx - set gpDebug=2
 -X N - set gpDebug=N (0..100)

 if gpDebug = 0, don't output any debug msgs
 if gpDebug > 0, output any "debug" msgs, or
 if gpDebug = N1, output "debug-N2" msgs, if N1 >= N2

See Globals: gpSysLog, gpFacility, gpVerbose, gpDebug

=internal-head4 fLog Examples:

 local tFile=${BASH_SOURCE##*/}
 local tFile=file.inc
 gpCmdName=program.sh

 fError -m "Message" -l $tFile:$LINENO -e 8

    Output: program.sh crit: Error: Message [file.inc:234](8)
    gErr=8
    exit 8

 fError -n -m "Message" -l $tFile:$LINENO -e 8

    Output: program.sh err: Error: Message [file.inc:234](8)
    gErr=8
    return 8

 fError -i [-n] -m "Message" -l $tFile:$LINENO -e 8

    Output: program.sh err: Internal: Error: Message [file.inc:234](8)
    gErr=8
    call: fComStackTrace
    return 8, if -n
    exit 8, if ! -n

 fLog -p err -m "Message" -l $tFile:$LINENO -e 8

    Output: program.sh err: Message [file.inc:234](8)
    gErr=8
    return 8

 fLog -p warning -m "Output if gpVerbose" -e 8

    if gpVerbose >= 1
    Output: program.sh warning: Message (8)
    gErr=8
    return 0

 fLog -p notice -m "Notice msg"  -e 8 -l $LINENO

    if gpVerbose >= 2
    Output: program.sh notice: Notice msg [234]
    gErr=0
    return 0

 fLog -p info -m "Info message"

    if gpVerbose >= 3
    Output: program.sh info: Info message
    gErr=0
    return 0

 fLog -p debug -m "Output only if $gpDebug > 0" -l $LINENO

    Output: program.sh debug: Output only if $gpDebug > 0 [234]
    return 0

 fLog -p debug-3 -m "Output only if $gpDebug >= 3" -l $LINENO

    Output: program.sh debug-3: Output only if $gpDebug >= 3 [234]
    return 0

=internal-cut
EOF
} # fLog

# --------------------------------
fError()
{
    local pMsg="Error"
    local pLine=""
    local pErr=1
    local pInt=""
    local pNoExit=0
    local tSrc=${BASH_SOURCE##*/}
    local tLevel=crit
    local OPTIND=1
    local OPTARG
    local tArg

    while getopts :m:l:e:in tArg; do
        case $tArg in
            m) pMsg="${OPTARG}" ;;
            l) pLine="${OPTARG}" ;;
            e) pErr="${OPTARG}" ;;
            i) pInt="-i" ;;
            n)
                pNoExit=1
                tLevel=err
                ;;
            :)
                echo "Internal: Value required for option: -$OPTARG [$tSrc:$LINENO]"
                fComStackTrace
                ;;
            \?)
                echo "Internal: Unknown option: $OPTARG [$tSrc:$LINENO]"
                fComStackTrace
                ;;
        esac
    done
    shift $((OPTIND - 1))
    if [ $# -ne 0 ]; then
        pMsg="$pMsg $*"
    fi
    if [ $# -ne 0 ]; then
        shift $#
    fi

    pMsg="Error: $pMsg"
    fLog -p $tLevel $pInt -m "$pMsg" -l $pLine -e $pErr

    if [ -n "$pInt" ] || [ $gpDebug -ge 2 ]; then
        fComStackTrace
    fi

    if [ -z "$pInt" ] && [ $pNoExit -eq 0 ]; then
        fUsage usage
        exit 1
    fi

    if [ $pNoExit -eq 0 ]; then
        exit 1
    fi
    return $gErr

    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fError -m "pMsg" [-l $LINENO] [-e pErr] [-i] [-n]

    If no -n
        if no -i
            fLog -p crit -m "Error: pMsg" -l pLine -e pErr
            fUsage short
            (exit 1)
        if -i
            fLog -p crit -m "Error: Internal: pMsg" -l pLine -e pErr
            fComStackTrace
            exit $pErr
    if -n
        if no -i
            fLog -p err -m "Error: pMsg" -l pLine -e pErr
        if -i
            fLog -p err -m "Error: Internal: pMsg" -l pLine -e pErr
            fComStackTrace
    if gpDebug >= 2
        fComStackTrace

Typical example:

    export gpCmd=git-proj-init

    testIt()
    {
        local tSrc=${BASH_SOURCE##*/}
        local Dir="foobar"

        if [ ! -w $tDir ]; then
            fError -n -i -m "Directory $tDir is not writable" -l $tSrc:$LINENO
        fi
        if [ ! -d $tDir ]; then
            fError -i -n -m "Directory $tDir not found" -l $tSrc:$LINENO
        fi
        fError -m "Too many errors" -l $tSrc:$LINENO
    }

Output:

   git-proj-init err: Error: Directory foobar is not writable [gpproj-com:1403]

   git-proj-init err: Error: Internal: Directory foobar not found [gpproj-com:1406]
   stack dump:

   git-proj-init crit: Error: Too many errors [gpproj-com:1408]
   Usage:

=internal-cut
EOF
} # fError

# --------------------
fComGit()
{
    local tSrc=${BASH_SOURCE##*/}
    local tOpt
    local tStatus

    tOpt=$(echo "$@" | sed 's/-/\-/g; s/"/\\\"/g')

    fLog -p info -l $tSrc:$LINENO -m "git $tOpt"
    git "$@"
    tStatus=$?
    if [ $tStatus -ne 0 ]; then
        fLog -p warning -m "Fix this then try again." -l $tSrc:$LINENO
    fi
    return $tStatus
    cat <<EOF
=internal-pod

=internal-for html <hr/>

=internal-head3 fComGit [pArgs...]

TODO: replace this with set/try/catch

This is a wrapper for "git". If gpVerbose is >= 2, the git call will
will be echoed. If the git call exits with a non-0, then an error
message is output and the script exits (if not called in a sub-shell).

This is used to make the code look "clean". You don't need to clutter
the code with if statements with every git command. I.e. this supports
the "design pattern" where, if a command exits, it is assumed all is
OK, because if it failed, the script would exit.

=internal-cut
EOF
} # fComGit

# --------------------------------
fComSetConfig()
{
    local pScope=""
    local pType=""
    declare -l pKey=""
    local pValue=""
    local pAdd=""
    local tSrc=${BASH_SOURCE##*/}
    local tTopDir
    local OPTIND=1
    local OPTARG
    local tArg
    while getopts :abf:gik:lLv: tArg; do
        # shellcheck disable=SC2213,SC2220
        case $tArg in
            f)
                if [ ! -f ${OPTARG} ]; then
                    fLog -p warning -m "Could not find ${OPTARG}, it will be created" -l "$tSrc:$LINENO"
                    touch ${OPTARG}
                    if [ ! -f ${OPTARG} ]; then
                        fError -n -m "Could not create ${OPTARG} in $PWD" -l "$tSrc:$LINENO"
                        return 1
                    fi
                fi
                ;;
            l)
                if ! fComMustBeInGitRepo $PWD; then
                    return 1
                fi
                ;;
            L)
                if ! fComMustBeInProjRepo $PWD; then
                    return 1
                fi
                tTopDir="$($cGetTopDir 2>&1)"
                ;;
        esac
        case $tArg in
            a) pAdd="--add" ;;
            b) pType="--bool" ;;
            f) pScope="--file ${OPTARG}" ;;
            g) pScope="--global" ;;
            i) pType="--int" ;;
            k) pKey="${OPTARG}" ;;
            l) pScope="--local" ;;
            L) pScope="--file $tTopDir/.gitproj" ;;
            v) pValue="${OPTARG}" ;;
            :)
                fError -n -i -m "fComSetConfig: Value required for option: $OPTARG" -l $tSrc:$LINENO
                return 1
                ;;
            \?)
                fError -n -i -m "fComSetConfig:: Unknown option: $OPTARG" -l $tSrc:$LINENO
                return 1
                ;;
        esac
    done
    if [ -z "$pKey" ]; then
        fError -n -i -m "fComSetConfig: missing key" -l $tSrc:$LINENO
        return 1
    fi
    if [ -z "$pValue" ]; then
        fError -n -i -m "fComSetConfig: missing value" -l $tSrc:$LINENO
        return 1
    fi

    git config $pScope $pAdd $pType $pKey "$pValue"
    return $?

    # ----------------------
    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComGetConfig

=internal-head4 Synopsis

    fComSetConfig -k pKey -v pValue [-g|-l|-L|-f pFile] [-b|-i] [-a]

=internal-head4 Description

=internal-head4 Options

Only one of these will be used: -g, -l, -L, -f. The last one in the
list of options will be used if there is more than one.  If none, then
-l is the default.

Only one or none of these can be specified: -b, -i. If more than one,
then the last one will be used.

=internal-over 4

=internal-item B<-k pKey>

This option is required.

=internal-item B<-v pValue>

This option is required.

=internal-item B<-g>

Write to ~/.gitconfig (--global)

=internal-item B<-l>

You must be in a git workspace for this option to work. (--local)

Write to GIT_DIR/.git/config

=internal-item B<-L>

Write to GIT_DIR/.gitproj

=internal-item B<-f pFile>

Write to pFile.

=internal-item B<-b>

The value will be normalized to "true" or "false".

=internal-item B<-i>

The value will be converted to a single "byte" value, based on
its suffix.

Valid suffixes: k, K, m, M, g, G

These will be converted to lowercase before being sent to "git config".

=internal-item B<-a>

Use this option if you need to specify a key with multiple values.
If not specified, then any matching key in the file, being written to,
will replace the value.

=internal-back

=internal-cut
EOF
} # fComSetConfig

# --------------------------------
fComGetConfig()
{
    local pScope=""
    local pType=""
    declare -l pKey=""
    local pDefault=""
    local pFilter=""
    local pExitOnError=0
    local tSrc=${BASH_SOURCE##*/}
    local tTopDir
    local tValue=""
    local tExit="-n"

    # Globals set by fComGetConfig
    gGetConfigValue=""
    gGetConfigOrigin="not-found"

    local OPTIND=1
    local OPTARG
    local tArg
    while getopts :bd:ef:gik:lLv: tArg; do
        # shellcheck disable=SC2213,SC2220
        case $tArg in
            f)
                if [ ! -f ${OPTARG} ]; then
                    fError $tExit -m "Could not find ${OPTARG}" -l "$tSrc:$LINENO"
                    return 1
                fi
                ;;
            l)
                if ! fComMustBeInGitRepo $PWD; then
                    return 1
                fi
                ;;
            L)
                if ! fComMustBeInProjRepo $PWD; then
                    return 1
                fi
                tTopDir="$($cGetTopDir 2>&1)"
                ;;
        esac
        case $tArg in
            b) pType="--bool" ;;
            d) pDefault="${OPTARG}" ;;
            e)
                pExitOnError=1
                tExit=""
                ;;
            f) pScope="--file ${OPTARG}" ;;
            g) pScope="--global" ;;
            i) pType="--int" ;;
            k) pKey="${OPTARG}" ;;
            l) pScope="--local" ;;
            L) pScope="--file $tTopDir/.gitproj" ;;
            v) pFilter="${OPTARG}" ;;
            :)
                fError $tExit -i -m "fComGetConfig: Value required for option: $OPTARG" -l $tSrc:$LINENO
                return 1
                ;;
            \?)
                fError $tExit -i -m "fComGetConfig: Unknown option: $OPTARG" -l $tSrc:LINENO
                return 1
                ;;
        esac
    done
    if [ -z "$pKey" ]; then
        fError $tExit -i -m "fComGetConfig: missing key" -l $tSrc:$LINENO
        return 1
    fi

    gGetConfigValue=$(git config --get --includes $pScope $pType $pKey $pFilter)

    if [ $pExitOnError -ne 0 ]; then
        if [ -z "$gGetConfigValue" ]; then
            fError -m "Unexpected: $pKey is not defined!" -l $tSrc:$LINENO
        fi

        if [ -n "$pDefault" ] && [ "$gGetConfigValue" = "$pDefault" ]; then
            fError -m "Unexpected: $pKey should not be set to: $gGetConfigValue" -l $tSrc:$LINENO
        fi

        if [ "$gGetConfigValue" = "TBD" ] || [ "$gGetConfigValue" = "not-defined" ]; then
            fError -m "Unexpected: $pKey should not be set to: $gGetConfigValue" -l $tSrc:$LINENO
        fi
    fi

    if [ -z "$gGetConfigValue" ] && [ -n "$pDefault" ]; then
        gGetConfigOrigin=default
        gGetConfigValue=$pDefault
        echo "$gGetConfigValue"
        return 0
    fi

    if [ -z "$gGetConfigValue" ]; then
        gGetConfigOrigin=not-found
        echo ""
        return 0
    fi

    gGetConfigOrigin=$(git config --show-origin --get --includes $pScope $pType $pKey $pFilter)
    gGetConfigOrigin=${gGetConfigOrigin% *}
    echo "$gGetConfigValue"
    return 0

    # ----------------------
    cat <<EOF >/dev/null
=internal-pod

=internal-for html <hr/>

=internal-head3 fComGetConfig

=internal-head4 Synopsis

    fComGetConfig -k pKey [-g|-l|-L|-f pFile] [-b|-i]
                  [-d pDefault] [-e] [-v pFilter]

=internal-head4 Description

If pKey is found, the last defined value is echoed. And the value is
saved in global gGetConfigValue. And the file location of the value is
saved in gGetConfigOrigin.

If the key is not found, and -d pDefault is set then gGetConfigValue
is set to pDefault, and gGetConfigOrigin is set to "default".

If the key is not found, and there is no -d option, then
gGetConfigValue is set to "", and gGetConfigOrigin is set to
"not-found", and "" is returned.

Command Returns

    0 - OK (even for no value found)
    1 - Errors

=internal-head4 Options

Only one of these will be used: -g, -l, -L, -f. The last one in the list
of options will be if there is more than one.  If none, then all the
config files will be used.

Only one or none of these can be specified: -b, -i. If more than one,
then the last one will be used.

=internal-over 4

=internal-item B<-k pKey>

This option is required. The last file were this key is found will be
used to set the value returned.

=internal-item B<-g>

Look in ~/.gitconfig Includes are followed so ~/gitproj.config.global
will be included.

=internal-item B<-l>

You must be in a git workspace for this option to work.

Look in GIT_DIR/.git/config Includes are followed so these files will
be included: GIT_DIR/.gitproj and
GIT_DIR/.git/config

=internal-item B<-f pFile>

This will look in pFile for pKey. Includes will be followed.

=internal-item B<-b>

The value will be normalized to "true" or "false".

Valid values in the files:

=internal-item B<-i>

The value will be converted to a single "byte" value, based on
its suffix.

Valid suffixes: k, m, g

Note: fComSetConf with -i will allow: [kKmMgG]. It will convert the
values to lower-case, when saving them in the config files.

=internal-item B<-d pDefault>

If pKey is not found, then pDefault will be returned, and gGetConfigOrigin
will be set to "default".

=internal-item B<-e>

If -e, then do some simple validation checks:

* if the pKey is not found, then fError will be called and the script
will terminate. I.e. the config var is required.

* or if the value for pKey is equal to pDefault, fError will be called
and the script will terminate. I.e. the config var is still set to
it's initial "undefined" value.

* or if the value for pKey is equal to "TBD" or "not-defined", fError
will be called and the script will terminate. I.e. the config var is
still set to it's initial "undefined" value.

Note: -e will only exit the whole script if it is not in a subshell.
I.e. don't use var=\$(cmd) style.

=internal-item B<-v pFilter>

pFilter is used as a pattern match for the value found. This is useful
for times when there could be duplicate pKeys with different
values. For example there could be multiple "include.path" keys.

=internal-back

=internal-cut
EOF
} # fComGetConfig

# --------------------------------
fComUnsetConfig()
{
    local pScope=""
    declare -l pKey=""
    local pAll="--unset"
    local tSrc=${BASH_SOURCE##*/}
    local tTopDir
    local OPTIND=1
    local OPTARG
    local tArg

    while getopts :af:gk:lL tArg; do
        # shellcheck disable=SC2213,SC2220
        case $tArg in
            f)
                if [ ! -f ${OPTARG} ]; then
                    fError -n -m "Could not find ${OPTARG}" -l "$tSrc:$LINENO"
                    return 1
                fi
                ;;
            l)
                if ! fComMustBeInGitRepo $PWD; then
                    return 1
                fi
                ;;
            L)
                if ! fComMustBeInProjRepo $PWD; then
                    return 1
                fi
                tTopDir="$($cGetTopDir 2>&1)"
                ;;
        esac
        case $tArg in
            a) pAll="--unset-all" ;;
            f) pScope="--file ${OPTARG}" ;;
            g) pScope=--global ;;
            k) pKey="${OPTARG}" ;;
            l) pScope=--local ;;
            L) pScope="--file $tTopDir/.gitproj" ;;
            :)
                fError -n -i -m "fComUnsetConfig: Value required for option: $OPTARG" -l $tSrc:$LINENO
                return 1
                ;;
            \?)
                fError -n -i -m "ComUnsetConfig: Unknown option: $OPTARG" -l $tSrc:$LINENO
                return 1
                ;;
        esac
    done

    if [ -z "$pKey" ]; then
        fError -n -i -m "fComUnsetConfig: missing key" -l $tSrc:$LINENO
        return 1
    fi

    git config $pAll $pScope $pKey
} # fComUnsetConfig

# --------------------
fComSaveVar2Config()
{
    local pFile="$1"
    declare -l pKey="$2"
    local pVarName="$3"
    local tSrc=${BASH_SOURCE##*/}
    local tValue

    tValue=$(eval echo \$$pVarName)
    if [ -z "$tValue" ]; then
        tValue="Undefined"
    fi
    # TODO: Use fComSetConfig? because it has more error handling.
    ##fComSetConfig -f $pFile -k $pKey -v "$tValue"
    git config --replace-all --file $pFile $pKey "$tValue"
    if [ $? -ne 0 ]; then
        fLog -i -p warning -m "There was problem saving $pVarName value to $pKey to file $pFile" -l $tSrc:$LINENO
        gErr=1
    fi
} # fComSaveVar2Config

# --------------------
fComCheckPkg()
{
    local pPkg=$1
    local tSrc=${BASH_SOURCE##*/}
    local tStatus

    dpkg -l $pPkg 2>/dev/null | grep -q "^ii *$pPkg"
    tStatus=$?
    fLog -p debug-2 -m "fComCheckPkg $pPkg $tStatus" -l $tSrc:$LINENO
    return $tStatus
} # fComCheckPkg

# --------------------
fComMkGitFlow()
{
    local tSrc=${BASH_SOURCE##*/}

    if [ "$gpGitFlow" = "false" ]; then
        return 0
    fi
    if ! fComCheckPkg git-flow; then
        echo
        fLog -p warning -m "git-flow is not installed" -l $tSrc:$LINENO
        gpGitFlow=false
        fComSetConfig -g -k gitproj.config.git-flow -v false
        return 1
    fi
    echo
    fLog -p notice -m "git flow init" -l $tSrc:$LINENO
    git flow init -d

    # Define the default prefixes
    fComSetConfig -g -k gitflow.branch.main -v main
    fComSetConfig -g -k gitflow.branch.master -v main
    fComSetConfig -g -k gitflow.branch.develop -v develop

    fComSetConfig -g -k gitflow.prefix.feature -v feature/
    fComSetConfig -g -k gitflow.prefix.bugfix -v bug/
    fComSetConfig -g -k gitflow.prefix.release -v release/
    fComSetConfig -g -k gitflow.prefix.hotfix -v hotfix/
    fComSetConfig -g -k gitflow.prefix.support -v support/

    return 0
} # fComMkGitFlow

# --------------------
fComMustBeInGitRepo()
{
    local pDir=$1
    local tSrc=${BASH_SOURCE##*/}

    if [ ! -d $pDir ]; then
        fError -n -m "$pDir does not exist." -l $tSrc:$LINENO
        return 1
    fi

    cd $pDir >/dev/null 2>&1
    if [ -d $PWD/.git ]; then
        cd - >/dev/null 2>&1
        return 0
    fi
    if $cGetTopDir >/dev/null 2>&1; then
        cd - >/dev/null 2>&1
        return 0
    fi
    cd - >/dev/null 2>&1
    fError -n -m "You must be in a git workspace for this command." -l $tSrc:$LINENO
    return 1
} # fComMustBeInGitRepo

# --------------------
fComMustNotBeInGit()
{
    local pDir=$1
    local tSrc=${BASH_SOURCE##*/}
    local tFound=0
    local tGitDir

    if [ ! -d $pDir ]; then
        fError -n -m "$pDir does not exist." -l $tSrc:$LINENO
        return 1
    fi
    cd $pDir >/dev/null 2>&1
    if $cGetTopDir >/dev/null 2>&1; then
        git branch
        tFound=1
    fi
    #    if [ $tFound -eq 0 ]; then
    #        tGitDir=$(find . -type d -name .git)
    #        if [ -n "$tGitDir" ]; then
    #            tFound=2
    #        fi
    #    fi
    cd - >/dev/null 2>&1
    if [ $tFound -ne 0 ]; then
        fError -n -m "$pDir must NOT be in a git repo" -l $tSrc:$LINENO
        return 1
    fi
    return 0
} # fComMustNotBeInGit

# --------------------
fComMustBeInProjRepo()
{
    local pDir=$1
    local tSrc=${BASH_SOURCE##*/}
    local tTopDir

    if ! fComMustBeInGitRepo $pDir; then
        return 1
    fi

    if [ ! -f $HOME/.gitconfig ]; then
        fError -n -m "Missing ~/.gitconfig, run 'git proj init'" -l $tSrc:$LINENO
        return 1
    fi

    tTopDir=$($cGetTopDir)

    if [ ! -f $tTopDir/.gitproj ]; then
        fError -n -m "This git workspace is not setup for gitproj." -l $tSrc:$LINENO
        if [ $gpDebug -ge 2 ]; then
            fComStackTrace
        fi
        return 1
    fi

    return 0
} # fComMustBeInProjRepo

fComAllMustBeReadable()
{
    local pDir=${1:-$PWD}
    local tSrc=${BASH_SOURCE##*/}
    local tFound

    tFound=$(find $pDir -type d \! -executable -print 2>/dev/null)
    if [ -n "$tFound" ]; then
        fError -n -m "All directories must be executable, under $pDir" -l $tSrc:$LINENO
        return 1
    fi

    tFound=$(find $pDir \! -readable -print 2>/dev/null)
    if [ -n "$tFound" ]; then
        fError -n -m "All directories and files must be readable, under $pDir" -l $tSrc:$LINENO
        return 1
    fi

    return 0
} # fComAllMustBeReadable

# --------------------
fComIsGitRemoteAvailable()
{
    # TODO not used yet
    local pLevel="${1:-debug}"
    local tSrc=${BASH_SOURCE##*/}
    local tOrigin

    tOrigin=$(fComGetConfig -k "remote.origin.url" -d "TBD")
    if [ "$tOrigin" = "TBD" ]; then
        fLog -p $pLevel -m "remote.origin.url is not defined." -l $tSrc:$LINENO
        return 1
    fi
    if git ls-remote $tOrigin >/dev/null 2>&1; then
        return 0
    fi
    fLog -p $pLevel -m "$tOrigin was not found. Try again after mounting it or run 'git proj remote' to change the remote.raw.dir location." -l $tSrc:$LINENO
    return 1
} # fComIsGitRemoteAvailable

# --------------------
fComIsRawRemoteAvailable()
{
    # TODO not used yet
    local pLevel="${1:-debug}"
    local tSrc=${BASH_SOURCE##*/}

    gpRemoteRawOrigin=$(fComGetConfig -k "gitproj.config.remote-raw-origin" -d "TBD")
    if [ "$gpRemoteRawOrigin" = "TBD" ]; then
        fLog -p $pLevel -m "gitproj.config.remote-raw-origin is not defined." -l $tSrc:$LINENO
        return 1
    fi
    # TODO: change this to use rclone
    if [ -d $gpRemoteRawOrigin ]; then
        return 0
    fi
    fLog -p $pLevel -m "$gpRemoteRawOrigin was not found. Try again after mounting it or run 'git proj remote' to change the remote.raw.dir location." -l $tSrc:$LINENO
    return 1
} # fComIsRawRemoteAvailable

# --------------------
fComIsRemoteMounted()
{
    local pLevel="${1:-debug}"
    local tSrc=${BASH_SOURCE##*/}

    # Assumes we are in a gitproj managed dir
    # And this assumes remotes are local disks

    gpRemoteRawOrigin=$(fComGetConfig -k "gitproj.config.remote-raw-origin" -d "TBD")
    if [ "$gpRemoteRawOrigin" = "TBD" ]; then
        fLog -p $pLevel -m "gitproj.config.remote-raw-origin is not defined." -l $tSrc:$LINENO
    fi
    if [ -d $gpRemoteRawOrigin ]; then
        return 0
    fi
    fLog -p $pLevel -m "$gpRemoteRawOrigin was not found. Try again after mounting it or run 'git proj remote' to change the remote.raw.dir location." -l $tSrc:$LINENO
    return 1
} # fComIsRemoteMounted

# --------------------
fComCheckDirSpace()
{
    local pMountDir=$1
    local pMinSpace=$2
    local tSrc=${BASH_SOURCE##*/}
    local tAvail
    local tGitSize
    local tRawSize
    local tDesiredSize

    tAvail=$(df -BM $pMountDir --output=avail | tail -n1)
    tAvail=${tAvail%M}

    tGitSize=$(du -sm $gpLocalTopDir/.git | awk '{print $1}')
    let tGitSize*=2

    tRawSize=$(du -sm $gpLocalTopDir/raw | awk '{print $1}')
    let tRawSize*=2

    let tDesiredSize=tGitSize+tRawSize
    let pMinSpace=pMinSpace/1024/1024

    if [ $tAvail -lt $tDesiredSize ] || [ $tAvail -lt $pMinSpace ]; then
        fLog -f -p warning -m "There may not be enough space at $pMountDir, which has ${tAvail}MB. .git and raw needs ${tDesiredSize}MB (or ${pMinSpace}MB)" -l $tSrc:$LINENO
        return 1
    fi

    return 0
} # fComCheckDirSpace

# --------------------
fComGetProjGlobals()
{
    local tSrc=${BASH_SOURCE##*/}

    # Assumes we in git dir, and one that is managed by gitproj.
    if ! fComMustBeInProjRepo $PWD; then
        fUsage usage
        exit 1
    fi

    gpSysLog=$(fComGetConfig -k "gitproj.config.syslog" -d "false" 2>&1)
    gpFacility=$(fComGetConfig -k "gitproj.config.facility" -d user)
    gpLocalTopDir=$($cGetTopDir 2>/dev/null)

    # Called this way, so that the -e option will exit script if errors
    fComGetConfig -k "gitproj.config.proj-name" -d "TBD" -e >/dev/null
    gpProjName=$gGetConfigValue

    fComGetConfig -k "gitproj.config.remote-raw-origin" -d "TBD" -e >/dev/null
    gpRemoteRawOrigin=$gGetConfigValue
    gpRemoteMinSpace=$(fComGetConfig -i -k "gitproj.config.remote-min-space" -d "2147483648" 2>&1) # 2g

    fComGetConfig -k "gitproj.config.local-status" -d "not-defined" -e >/dev/null
    gpLocalStatus=$gGetConfigValue

    fComGetConfig -k "gitproj.config.remote-status" -d "not-defined" -e >/dev/null
    gpRemoteStatus=$gGetConfigValue

    fComGetVer

    return 0
} # fComGetProjGlobals

# --------------------
fComMkPreCommit()
{
    local tSrc=${BASH_SOURCE##*/}
    local tVerbose=""

    if ! fComMustBeInGitRepo $PWD; then
        return 1
    fi

    cd $gpLocalTopDir >/dev/null 2>&1
    if [ $gpVerbose -ge 2 ]; then
        tVerbose='-v'
    fi

    if [ ! -f ~/.pre-commit ]; then
        fLog -f -p notice -m "Update ~/.pre-commit from $gpDoc/hooks/pre-commit" -l $tSrc:$LINENO
        cp -p $tVerbose $gpDoc/hooks/pre-commit ~/.pre-commit
    fi

    if [ ! -f .pre-commit ]; then
        fLog -f -p notice -m "Update $gpProjName/.pre-commit from ~/.pre-commit and add to git" -l $tSrc:$LINENO
        cp -p $tVerbose ~/.pre-commit .
        git add .pre-commit
    fi

    if [ ! -f .git/hooks/pre-commit ]; then
        fLog -f -p notice -m "Update .git/hooks/pre-commit from $gpProjName/.pre-commit" -l $tSrc:$LINENO
        cp -p $tVerbose .pre-commit .git/hooks/pre-commit
    fi
    if [ ! -x ./.git/hooks/pre-commit ]; then
        chmod a+rx ./.git/hooks/pre-commit
    fi

    return 0
} # fComMkPreCommit

fComGetVer()
{
    local tSrc=${BASH_SOURCE##*/}
    local tInsVer
    local tInsMajor
    local tInsMinor
    local tLocVer
    local tLocMajor
    local tLocMinor
    local tTrim

    if ! fComMustBeInProjRepo $PWD; then
        fError -m "Quitting" -l $tSrc:$LINENO
    fi

    # Get the expected Version for the "current" project
    tLocVer=$(fComGetConfig -k "gitproj.config.ver")
    if [ -z "$tLocVer" ]; then
        gpVer=$cGitProjVersion
        fLog -i -p warning -m "gitproj.config.ver was not found. Setting it to: $gpVer" -l "$tSrc:$LINENO"
        fComSetConfig -L -k "gitproj.config.ver" -v "$gpVer"
        fComSetConfig -l -k "gitproj.config.ver" -v "$gpVer"
    fi
    gpVer=$tLocVer

    if [ "$gpVer" = "$cGitProjVersion" ]; then
        return 0
    fi

    # Prep: tInsVer. Only the first two # are important
    tTrim=${cGitProjVersion#*.*.}
    tInsVer=${cGitProjVersion%.$tTrim}
    tInsMajor=${tInsVer%.*}
    tInsMinor=${tInsVer#*.}

    # Prep: tLocVer. Only the first two # are important
    tTrim=${gpVer#*.*.}
    tLocVer=${gpVer%.$tTrim}
    tLocMajor=${tLocVer%.*}
    tLocMinor=${tLocVer#*.}

    if [ $tLocMajor -gt $tInsMajor ]; then
        fError -n -m "Expected version $tLocVer The installed version $cGitProjVersion needs to be upgraded to $tLocVer or greater." -l $tSrc:$LINENO
        return 1
    fi

    if [ $tLocMajor -lt $tInsMajor ]; then
        fError -n -m "Expected version $tLocVer  The project needs to be upgraded to the installed version $cGitProjVersion" -l $tSrc:$LINENO
        return 1
    fi

    if [ $tLocMinor -ne $tInsMinor ]; then
        fLog -p warning -m "Minor version difference. Expected version $tLocVer, but installed version is $tInsVer" -l $tSrc:$LINENO
        return 0
    fi

    return 0
} # fComGetVer

fComFmt()
{
    # Call fmt, setting the width based on the the terminal COLUMNS
    # width.  Set to COLUMNS - 5, but no less than 30 and no more than
    # 130 characters wide.

    local tWidth

    if [ -z "$COLUMNS" ]; then
        COLUMNS=80
    fi
    let tWidth=COLUMNS-5
    if [ $tWidth -lt 30 ]; then
        tWidth=30
    fi
    if [ $tWidth -gt 130 ]; then
        tWidth=130
    fi
    fmt -w $tWidth
} # fComFmt

# ========================================
fComSetGlobals
